# -*- coding: utf-8 -*-
"""
Loads configuration.
"""

from __future__ import absolute_import, unicode_literals

import glob
import io
import os

import yaml

from polarion_docstrings import utils

PROJECT_CONF_DIRS = ("conf", os.curdir)
PROJECT_CONF = "polarion_tools*.yaml"


def get_config():
    """Loads configuration from project config file."""
    config_settings = {}

    project_root = utils.find_vcs_root(os.curdir)
    if project_root is None:
        return config_settings

    for conf_dir in PROJECT_CONF_DIRS:
        conf_dir = conf_dir.lstrip("./")
        joined_dir = os.path.join(project_root, conf_dir) if conf_dir else project_root
        joined_glob = os.path.join(joined_dir, PROJECT_CONF)
        conf_files = glob.glob(joined_glob)
        if conf_files:
            break
    else:
        conf_files = []

    for conf_file in conf_files:
        try:
            with io.open(conf_file, encoding="utf-8") as input_file:
                loaded_settings = yaml.load(input_file)
        except EnvironmentError:
            pass
        else:
            config_settings.update(loaded_settings)

    return config_settings



# encoding: utf-8
# pylint: disable=missing-docstring,no-self-use

from __future__ import unicode_literals

import pytest


class TestClassFoo(object):
    def test_in_class_no_docstring(self):
        pass

    def test_in_class_no_polarion(self):
        """FOO"""
        pass

    @pytest.mark.skip
    def test_in_class_polarion(self):
        """FOO

        Polarion:
            assignee: mkourim
            casecomponent: nonexistent
            testSteps:
                1. Step with really long description
                   that doesn't fit into one line
                2. Do that
            expectedResults:
                1. Success outcome with really long description
                   that doesn't fit into one line
                2. second
            caseimportance: low
            title: Some test with really long description
                   that doesn't fit into one line
            setup: Do this:
                   - first thing
                   - second thing
            teardown: Tear it down
            caselevel: level1
            caseautomation: automated
            linkedWorkItems: FOO, BAR
            foo: this is an unknown field
            description: ignored

        This is not included.
        """
        pass


@pytest.mark.skip
def test_annotated_no_docstring():
    pass


def test_standalone_no_docstring():
    pass


@pytest.mark.skip
def test_annotated_no_polarion():
    """FOO"""
    pass


@pytest.mark.skip
def test_annotated_polarion():

    """FOO

    Polarion:
        assignee: mkourim
        initialEstimate: 1/4
    """
    pass



# encoding: utf-8
# pylint: disable=missing-docstring,redefined-outer-name,no-self-use

from __future__ import unicode_literals

import os

import pytest

DATA_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


@pytest.fixture(scope="module")
def source_file():
    return os.path.join(DATA_PATH, "polarion_docstrings.py")



# -*- coding: utf-8 -*-
"""
Utility functions.
"""

from __future__ import absolute_import, unicode_literals

import os


def find_in_parent_dirs(path, filenames):
    """Searches up from a given path to find files."""
    prev, path = None, os.path.abspath(path)
    while prev != path:
        if any(os.path.exists(os.path.join(path, d)) for d in filenames):
            return path
        prev, path = path, os.path.abspath(os.path.join(path, os.pardir))
    return None


def find_tests_marker(path, markers=(".polarion_tests",)):
    """Searches up from a given path to find the tests markers."""
    return find_in_parent_dirs(path, markers)


def find_vcs_root(path, dirs=(".git",)):
    """Searches up from a given path to find the project root."""
    return find_in_parent_dirs(path, dirs)



# -*- coding: utf-8 -*-
"""
Parses Polarion docstring.
"""

from __future__ import absolute_import, unicode_literals

import ast
from collections import namedtuple

FORMATED_KEYS = ("setup", "teardown")

DocstringsRecord = namedtuple("DocstringsRecord", "lineno column value")


def get_section_start(doc_list, section):
    """Finds the line with "section" (e.g. "Polarion", "testSteps", etc.)."""
    section = "{}:".format(section)
    for index, line in enumerate(doc_list):
        if section != line.strip():
            continue
        indent = len(line) - len(line.lstrip(" "))
        return index + 1, indent
    return None, None


# pylint: disable=too-few-public-methods,invalid-name
class DOCSTRING_SECTIONS(object):  # noqa
    """Valid sections in Polarion docstring."""

    polarion = "Polarion"
    steps = "testSteps"
    results = "expectedResults"


class DocstringParser(object):
    """Parser for single docstring."""

    SECTIONS = DOCSTRING_SECTIONS

    def __init__(self, docstring):
        self.docstring = docstring

    @staticmethod
    def _get_key_value(line):
        """Gets the key and value out of docstring line."""
        data = line.split(":")
        if len(data) == 1:
            data.append("")

        key = data[0].strip()

        value = ":".join(data[1:]).strip()
        if value == "None":
            value = None

        return key, value

    # pylint: disable=too-many-locals
    def lines_to_dict(self, lines, start=0, lineno_offset=0, stop=None):
        """Creates dictionary out of docstring lines.

        Includes column and line number info for each record.
        """
        if stop:
            lines = lines[start:stop]
        else:
            lines = lines[start:]

        lines_dict = {}
        indent = len(lines[0]) - len(lines[0].lstrip(" "))
        prev_key = None
        for num, line in enumerate(lines, 1):
            line_stripped = line.strip()
            if not line_stripped:
                break

            curr_indent = len(line) - len(line.lstrip(" "))

            if curr_indent < indent:
                break

            # check if the line should be appended to previous key
            first_word = line_stripped.split(" ")[0] or line_stripped
            if prev_key and curr_indent > indent and first_word[-1] != ":":
                sep = "\n" if prev_key in FORMATED_KEYS else " "
                prev_lineno, prev_indent, prev_value = lines_dict[prev_key]
                lines_dict[prev_key] = DocstringsRecord(
                    prev_lineno, prev_indent, "{}{}{}".format(prev_value, sep, line_stripped)
                )
                continue
            else:
                prev_key = None

            if curr_indent > indent:
                continue

            key, value = self._get_key_value(line)
            lines_dict[key] = DocstringsRecord(num + lineno_offset, indent, value)
            prev_key = key

        return lines_dict

    @staticmethod
    def lines_to_list(lines, start=0, lineno_offset=0, stop=None):
        """Creates list out of docstring lines.

        Includes column and line number info for each line.
        """
        if stop:
            lines = lines[start:stop]
        else:
            lines = lines[start:]

        lines_list = []
        indent = len(lines[0]) - len(lines[0].lstrip(" "))
        for num, line in enumerate(lines, 1):
            line_stripped = line.strip()
            if not line_stripped:
                continue
            curr_indent = len(line) - len(line.lstrip(" "))
            if curr_indent < indent:
                break

            # check if the line should be appended to previous key
            first_word = line_stripped.split(" ")[0] or line_stripped
            if curr_indent > indent and first_word[-1] != ":":
                prev_lineno, prev_indent, prev_value = lines_list.pop()
                lines_list.append(
                    DocstringsRecord(
                        prev_lineno, prev_indent, "{} {}".format(prev_value, line_stripped)
                    )
                )
                continue

            if curr_indent > indent:
                continue
            lines_list.append(DocstringsRecord(num + lineno_offset, indent, line_stripped))

        return lines_list

    def strip_polarion_data(self):
        """Removes the Polarion section from the docstring."""
        docstring_list = self.docstring.split("\n")
        new_docstring_list = []
        indent = 0
        polarion_section = "{}:".format(self.SECTIONS.polarion)
        for line in docstring_list:
            if line.strip() == polarion_section:
                indent = len(line) - len(line.lstrip(" "))
                continue
            if indent:
                curr_indent = len(line) - len(line.lstrip(" "))
                if curr_indent > indent:
                    continue
            new_docstring_list.append(line)
        new_docstring = "\n".join(new_docstring_list).replace("\n\n", "\n")
        return new_docstring.strip()

    def parse(self):
        """Parses docstring to dictionary. E.g.:

        Polarion:
            assignee: mkourim
            casecomponent: nonexistent
            testSteps:
                1. Step with really long description
                   that doesn't fit into one line
                2. Do that
            expectedResults:
                1. Success outcome with really long description
                   that doesn't fit into one line
                2. 2
            caseimportance: low
            title: Some test with really long description
                   that doesn't fit into one line
            setup: Do this:
                   - first thing
                   - second thing

        This is not included.
        """
        doc_list = self.docstring.split("\n")

        polarion_start, __ = get_section_start(doc_list, self.SECTIONS.polarion)
        if not polarion_start:
            return None

        docstring_dict = self.lines_to_dict(doc_list, start=polarion_start)
        if self.SECTIONS.steps in docstring_dict and docstring_dict[self.SECTIONS.steps][2]:
            steps_start, __ = get_section_start(doc_list, self.SECTIONS.steps)
            steps_list = self.lines_to_list(
                doc_list, start=steps_start, lineno_offset=steps_start - polarion_start
            )
            docstring_dict[self.SECTIONS.steps] = steps_list
        if self.SECTIONS.results in docstring_dict and docstring_dict[self.SECTIONS.results][2]:
            results_start, __ = get_section_start(doc_list, self.SECTIONS.results)
            results_list = self.lines_to_list(
                doc_list, start=results_start, lineno_offset=results_start - polarion_start
            )
            docstring_dict[self.SECTIONS.results] = results_list

        return docstring_dict


class FileParser(object):
    """Parser for whole file."""

    SECTIONS = DOCSTRING_SECTIONS

    def __init__(self, filename, tree=None, tests_only=True):
        self.filename = filename
        self.tests_only = tests_only
        self.tree = tree or self.get_tree()

    def get_tree(self):
        """Returns ast tree."""
        with open(self.filename) as infile:
            source = infile.read()

        tree = ast.parse(source, filename=self.filename)
        return tree

    @staticmethod
    def get_docstring_from_func(func_node):
        """Gets docstring from function definition."""
        docstring = None
        try:
            if (
                func_node.body
                and isinstance(func_node.body[0], ast.Expr)
                and isinstance(func_node.body[0].value, ast.Str)
            ):
                docstring = func_node.body[0].value.s
        # pylint: disable=broad-except
        except Exception:
            return None

        # for Python2
        try:
            docstring = docstring.decode("utf-8")
        except AttributeError:
            pass

        return docstring

    def get_docstrings(self):
        """Returns parsed Polarion docstrings present in the python source file."""
        docstrings = []
        for node in ast.walk(self.tree):
            if not isinstance(node, ast.FunctionDef):
                continue

            if self.tests_only and node.name.find("test_") != 0:
                continue

            docstring = self.get_docstring_from_func(node)

            # test doesn't have docstring, i.e. it's missing also the Polarion section
            if not docstring:
                docstrings.append(DocstringsRecord(node.body[0].lineno - 1, node.col_offset, {}))
                continue

            doc_list = docstring.split("\n")
            docstring_start = node.body[0].lineno - len(doc_list)
            polarion_start, polarion_column = get_section_start(doc_list, self.SECTIONS.polarion)

            if not polarion_start:
                # docstring is missing the Polarion section
                docstrings.append(DocstringsRecord(docstring_start + 1, node.col_offset + 4, {}))
                continue

            polarion_offset = docstring_start + polarion_start
            docstrings.append(
                DocstringsRecord(polarion_offset, polarion_column, parse_docstring(docstring))
            )

        return docstrings


# convenience functions


def parse_docstring(docstring):
    """Parses docstring to dictionary."""
    return DocstringParser(docstring).parse()


def strip_polarion_data(docstring):
    """Removes the Polarion section from the docstring."""
    return DocstringParser(docstring).strip_polarion_data()


def get_docstrings_in_file(filename, tree=None, tests_only=True):
    """Returns parsed Polarion docstrings present in the python source file."""
    return FileParser(filename, tree=tree, tests_only=tests_only).get_docstrings()



# -*- coding: utf-8 -*-
"""
Checks Polarion docstrings.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import sys
from collections import namedtuple

from pkg_resources import get_distribution

from polarion_docstrings import configuration, parser, utils

DocstringsError = namedtuple("DocstringsError", "lineno column message func")
FieldRecord = namedtuple("FieldRecord", "lineno column field")
ValidatedDocstring = namedtuple("ValidatedDocstring", "unknown invalid missing markers ignored")


# pylint: disable=too-many-instance-attributes
class DocstringsChecker(object):
    """Checker for Polarion docstrings."""

    IGNORED_FIELD = "P664"
    MISSING_SECTION = "P665"
    UNKNOWN_FIELD = "P666"
    INVALID_VALUE = "P667"
    MARKER_FIELD = "P668"
    MISSING_FIELD = "P669"

    def __init__(self, tree, filename, config, func):
        self.tree = tree
        self.filename = filename
        self.config = config
        self.func = func

    @staticmethod
    def validate_value(docstring_dict, key, valid_values):
        record = docstring_dict.get(key)
        if record is not None:
            return record.value in valid_values[key]
        return True

    @staticmethod
    def get_unknown_fields(docstring_dict, known_fields):
        unknown = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key in docstring_dict
            if key not in known_fields
        ]
        return unknown

    @classmethod
    def get_invalid_fields(cls, docstring_dict, valid_values):
        results = {
            key: cls.validate_value(docstring_dict, key, valid_values) for key in valid_values
        }
        invalid = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key, result in results.items()
            if not result
        ]
        return invalid

    @staticmethod
    def get_missing_fields(docstring_dict, required_keys):
        if not required_keys:
            return []
        missing = [key for key in required_keys if key not in docstring_dict]
        return missing

    @staticmethod
    def get_markers_fields(docstring_dict, marker_fields):
        if not marker_fields:
            return []
        markers = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key in marker_fields
            if key in docstring_dict
        ]
        return markers

    @staticmethod
    def get_ignored_fields(docstring_dict, ignored_fields):
        if not ignored_fields:
            return []
        ignored = [
            FieldRecord(docstring_dict[key].lineno, docstring_dict[key].column, key)
            for key in ignored_fields
            if key in docstring_dict
        ]
        return ignored

    def validate_docstring(self, docstring_dict):
        """Returns tuple with lists of problematic fields."""
        cfg_docstrings = self.config["docstrings"]
        unknown = self.get_unknown_fields(docstring_dict, self.config.get("default_fields"))
        invalid = self.get_invalid_fields(docstring_dict, cfg_docstrings.get("valid_values"))
        missing = self.get_missing_fields(docstring_dict, cfg_docstrings.get("required_fields"))
        markers = self.get_markers_fields(docstring_dict, cfg_docstrings.get("marker_fields"))
        ignored = self.get_ignored_fields(docstring_dict, cfg_docstrings.get("ignored_fields"))
        return ValidatedDocstring(unknown, invalid, missing, markers, ignored)

    # pylint:disable=too-many-locals
    def get_fields_errors(self, validated_docstring, docstring_dict, lineno=0, column=0):
        """Produces fields errors for the flake8 checker."""
        errors = []
        cfg_docstrings = self.config["docstrings"]
        marker_fields = cfg_docstrings.get("marker_fields") or {}
        ignored_fields = cfg_docstrings.get("ignored_fields") or {}

        for num, col, field in validated_docstring.unknown:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Unknown field "{}"'.format(self.UNKNOWN_FIELD, field),
                    self.func,
                )
            )
        for num, col, field in validated_docstring.invalid:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Invalid value "{}" of the "{}" field'.format(
                        self.INVALID_VALUE, docstring_dict[field].value, field
                    ),
                    self.func,
                )
            )
        for num, col, field in validated_docstring.markers:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Field "{}" should be handled by the "{}" marker'.format(
                        self.MARKER_FIELD, field, marker_fields.get(field)
                    ),
                    self.func,
                )
            )
        for num, col, field in validated_docstring.ignored:
            errors.append(
                DocstringsError(
                    lineno + num,
                    col,
                    '{} Ignoring field "{}": {}'.format(
                        self.IGNORED_FIELD, field, ignored_fields.get(field)
                    ),
                    self.func,
                )
            )
        for field in validated_docstring.missing:
            errors.append(
                DocstringsError(
                    lineno,
                    column,
                    '{} Missing required field "{}"'.format(self.MISSING_FIELD, field),
                    self.func,
                )
            )

        if errors:
            errors = sorted(errors, key=lambda tup: tup[0])
        return errors

    @staticmethod
    def print_errors(errors):
        """Prints errors without using flake8."""
        for err in errors:
            print("line: {}:{}: {}".format(err.lineno, err.column, err.message), file=sys.stderr)

    def check_docstrings(self, docstrings):
        """Runs checks on each docstring."""
        errors = []
        for record in docstrings:
            if record.value:
                valdoc = self.validate_docstring(record.value)
                errors.extend(
                    self.get_fields_errors(valdoc, record.value, record.lineno, record.column)
                )
            else:
                errors.append(
                    DocstringsError(
                        record.lineno,
                        record.column,
                        '{} Missing "Polarion" section'.format(self.MISSING_SECTION),
                        self.func,
                    )
                )
        return errors

    def run_checks(self):
        """Checks docstrings in python source file."""
        docstrings = parser.get_docstrings_in_file(self.filename, tree=self.tree)
        errors = self.check_docstrings(docstrings)
        return errors

    def is_for_check(self):
        """Decides if the file should be checked."""
        # check only if configuration is valid
        cfg_valid = self.config.get("docstrings") or {}
        cfg_valid = cfg_valid.get("valid_values")
        if not (cfg_valid and self.config.get("default_fields")):
            return False

        abs_filename = os.path.abspath(self.filename)
        head, tail = os.path.split(abs_filename)

        # check only test files under polarion tests path
        if not (tail.startswith("test_") and utils.find_tests_marker(head or ".")):
            return False

        return True

    def get_errors(self):
        """Get errors in docstrings in python source file."""
        if not self.is_for_check():
            return []

        return self.run_checks()


def polarion_checks492(tree, filename):
    """The flake8 entry point."""
    config = configuration.get_config()
    return DocstringsChecker(tree, filename, config, polarion_checks492).get_errors()


try:
    # __package__ is not in python 2.7
    __version__ = get_distribution(__name__.split(".")[0]).version
# pylint: disable=broad-except
except Exception:
    # package is not installed
    __version__ = "0.0"

polarion_checks492.name = "polarion_checks"
polarion_checks492.version = __version__



# encoding: utf-8
# pylint: disable=missing-docstring

from __future__ import unicode_literals

import io
import os

import yaml

from polarion_docstrings import checker

CONFIG_TEMPLATE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), os.pardir, "polarion_tools.yaml.template"
)

RESULTS = [
    (51, 0, 'P665 Missing "Polarion" section'),
    (55, 0, 'P665 Missing "Polarion" section'),
    (61, 4, 'P665 Missing "Polarion" section'),
    (10, 4, 'P665 Missing "Polarion" section'),
    (14, 8, 'P665 Missing "Polarion" section'),
    (21, 8, 'P669 Missing required field "initialEstimate"'),
    (23, 12, 'P667 Invalid value "nonexistent" of the "casecomponent" field'),
    (39, 12, 'P667 Invalid value "level1" of the "caselevel" field'),
    (39, 12, 'P668 Field "caselevel" should be handled by the "@pytest.mark.tier" marker'),
    (40, 12, 'P668 Field "caseautomation" should be handled by the "@pytest.mark.manual" marker'),
    (
        41,
        12,
        'P668 Field "linkedWorkItems" should be handled by the "@pytest.mark.requirements" marker',
    ),
    (42, 12, 'P666 Unknown field "foo"'),
    (43, 12, 'P664 Ignoring field "description": use test docstring instead'),
]


def _strip_func(errors):
    return [(lineno, col, msg) for lineno, col, msg, __ in errors]


def test_checker(source_file):
    with io.open(CONFIG_TEMPLATE, encoding="utf-8") as config_file:
        config = yaml.load(config_file)
    errors = checker.DocstringsChecker(
        None, source_file, config, checker.polarion_checks492
    ).run_checks()
    errors = _strip_func(errors)
    assert len(errors) == len(RESULTS)
    assert errors == RESULTS



# encoding: utf-8
# pylint: disable=missing-docstring

from __future__ import unicode_literals

from polarion_docstrings import parser

RESULTS = [
    (51, 0, {}),
    (55, 0, {}),
    (61, 4, {}),
    (70, 4, {"assignee": (1, 8, "mkourim"), "initialEstimate": (2, 8, "1/4")}),
    (10, 4, {}),
    (14, 8, {}),
    (
        21,
        8,
        {
            "assignee": (1, 12, "mkourim"),
            "caseautomation": (19, 12, "automated"),
            "casecomponent": (2, 12, "nonexistent"),
            "caseimportance": (11, 12, "low"),
            "caselevel": (18, 12, "level1"),
            "expectedResults": [
                (
                    8,
                    16,
                    "1. Success outcome with really long description that doesn't "
                    "fit into one line",
                ),
                (10, 16, "2. second"),
            ],
            "foo": (21, 12, "this is an unknown field"),
            "linkedWorkItems": (20, 12, "FOO, BAR"),
            "setup": (14, 12, "Do this:\n- first thing\n- second thing"),
            "teardown": (17, 12, "Tear it down"),
            "testSteps": [
                (4, 16, "1. Step with really long description that doesn't fit into one line"),
                (6, 16, "2. Do that"),
            ],
            "title": (
                12,
                12,
                "Some test with really long description that doesn't fit into one line",
            ),
            "description": (22, 12, "ignored"),
        },
    ),
]


def test_parser(source_file):
    docstrings = parser.get_docstrings_in_file(source_file)
    assert len(docstrings) == len(RESULTS)
    assert docstrings == RESULTS
